#include  <Uefi.h>
#include  <Library/UefiLib.h>
#include  <Library/ShellCEntryLib.h>
#include <Protocol/BlockIo.h>
#include <Library/MemoryAllocationLib.h>
#include <string.h>
#include <Library/ShellLib.h>
#include  <Library/PrintLib.h>

#define DEBUGMODE 0
#define TEMPSTRINGLENGTH 200

extern EFI_BOOT_SERVICES         *gBS;
extern EFI_SYSTEM_TABLE			 *gST;
extern EFI_RUNTIME_SERVICES 	 *gRT;

// 用于  8MB PSRAM
#define MAXPSRAMADDRESS   8 * 1024 * 1024
// 用于 256MB Disk
#define MAXDISKADDRESS   256 * 1024 * 1024

UINT8 *SIGNATURE="DATTRANSFTERDISKWWW.LAB-Z.COM";
UINT8 *Block;

struct TranHeader
{
	UINT32 Flag;
	UINT32 Monot;
	UINT32 Size;
	UINT32 Counter;
	UINT16 FilenameLength;
	UINT8  Filename[494];
} tr;

INTN
EFIAPI
main (
    IN UINTN Argc,
    IN CHAR8 **Argv
)
{
	EFI_STATUS				Status;
	UINTN					HandleCount,HandleIndex,i;
	EFI_HANDLE              *BlockControllerHandles = NULL;
	EFI_BLOCK_IO_PROTOCOL   *BlockIo;
	BOOLEAN					Found=FALSE;
	EFI_FILE_HANDLE   		FileHandle;

	//找到全部有 BlockIo Protocol 的Device
	Status = gBS->LocateHandleBuffer(
	             ByProtocol,
	             &gEfiBlockIoProtocolGuid,
	             NULL,
	             &HandleCount,
	             &BlockControllerHandles);
	if (!EFI_ERROR(Status))
	{
		//逐个打开
		for (HandleIndex = 0; HandleIndex < HandleCount; HandleIndex++)
		{
			/*打开EFI_BLOCK_IO_PROTOCOL  */
			Status = gBS->HandleProtocol(
			             BlockControllerHandles[HandleIndex],
			             &gEfiBlockIoProtocolGuid,
			             (VOID**)&BlockIo);
			if (DEBUGMODE)
			{
				//显示信息
				Print(L"-->[Device]<--:%d\n",HandleIndex);
				Print(L"MediaId       :%0x\n",BlockIo->Media->MediaId);
				Print(L"RemovableMedia:%0x\n",BlockIo->Media->RemovableMedia);
				Print(L"MediaPresent  :%0x\n",BlockIo->Media->MediaPresent);
				Print(L"ReadOnly      :%0x\n",BlockIo->Media->ReadOnly);
				Print(L"WriteCaching  :%0x\n",BlockIo->Media->WriteCaching);
				Print(L"BlockSize     :%0x\n",BlockIo->Media->BlockSize);
				Print(L"IoAlign       :%0x\n",BlockIo->Media->IoAlign);
				Print(L"LastBlock     :%0x\n",BlockIo->Media->LastBlock);
				Print(L"LogicalPartition :%0x\n",BlockIo->Media->LogicalPartition);
				Print(L"LowestAlignedLba :%0x\n",BlockIo->Media->LowestAlignedLba);
				Print(L"LogicalBlocksPerPhysicalBlock   : %0x\n",
				      BlockIo->Media->LogicalBlocksPerPhysicalBlock);
				Print(L"OptimalTransferLengthGranularity: %0x\n",
				      BlockIo->Media->OptimalTransferLengthGranularity);
			}

			// 判断当前的 BlockIo 必须满足以下条件
			// 1. 必须是 RemovableMedia
			// 2. LastBlock 必须为0x7FFF
			if ((BlockIo->Media->RemovableMedia!=1)||(BlockIo->Media->LastBlock!=0x7FFFF))
			{
				if (DEBUGMODE)
				{
					Print(L"Break \n");
				}
				continue;
			}

			// 分配一个 BlockSize 大小的内存
			Block = AllocateZeroPool(BlockIo->Media->BlockSize);
			// 签名检查，用于判定设备是否为 DataTransfer
			BlockIo->ReadBlocks(
			    BlockIo,
			    BlockIo->Media->MediaId,
			    0x004, //0x800
			    BlockIo->Media->BlockSize,
			    Block);
			// 检查签名
			if (memcmp(SIGNATURE,Block,sizeof(SIGNATURE))!=0)
			{
				Print(L"Coundn't find Data Transfer\n");
				continue;
			}
			else
			{
				Found=TRUE;
				//Print(L"Found Data Transfer！\n");
				break;
			}

			FreePool(Block);
		}	//for (HandleIndex = 0;

		if (Found)
		{
			Print(L"We found a DataTransfer Disk\n");
		}
		else
		{
			// 系统中无法找到 DataTransfer 设备
			Print(L"We coundn't find DataTransfer Disk in system\n");
			return EFI_SUCCESS;
		}

		// 检查 Busy 位
		BlockIo->ReadBlocks(
		    BlockIo,
		    BlockIo->Media->MediaId,
		    0x0C, //0x1800
		    BlockIo->Media->BlockSize,
		    Block);
		// 如果当前设备 Busy 中，那么退出
		if ((Block[0] & 1) == 0)
		{
			Print(L"Device busy, please try again later!\n");
			return EFI_SUCCESS;
		}


		tr.Flag =    Block[0] + (Block[1] << 8) + (Block[2] << 16) + (Block[3] << 24);
		tr.Monot =   Block[4] + (Block[5] << 8) + (Block[6] << 16) + (Block[7] << 24);
		tr.Size =    Block[8] + (Block[9] << 8) + (Block[10] << 16) + (Block[11] << 24);
		tr.Counter = Block[12] + (Block[13] << 8) + (Block[14] << 16) + (Block[15] << 24);
		tr.FilenameLength = Block[16] + (Block[17] << 8);
		for (i = 0; i < tr.FilenameLength; i++)
		{
			tr.Filename[i] = Block[18 + i];
		}

		Print(L"Filename %a\n",tr.Filename);
		Print(L"Filesize %d\n",tr.Size);

		// 生成一个文件名
		// 1.在当前文件名中查找盘符
		for (i=0; i<AsciiStrLen(Argv[0]); i++)
		{
			if (Argv[0][i]==':')
			{
				if (DEBUGMODE)
				{
					Print(L"[%d]\n",i);
				}
				break;
			}
		}

		if (DEBUGMODE)
		{
			Print(L"Argv[0]==[%a]\n",Argv[0]);
		}

		CHAR16	DriveLetter[TEMPSTRINGLENGTH];
		CHAR16	tmpStr[TEMPSTRINGLENGTH];

		// 2. 取出盘符字符串
		// 将  Ascii 转为 Unicode
		UnicodeSPrintAsciiFormat(
		    &tmpStr[0],
		    TEMPSTRINGLENGTH,
		    Argv[0]);
		StrnCpyS(DriveLetter,TEMPSTRINGLENGTH,tmpStr,i+2);

		if (DEBUGMODE)
		{
			Print(L">[%s]\n",tmpStr);
			Print(L">[%s]\n",DriveLetter);
		}

		// 3. 将设备存储的文件名转化为 Unicode
		UnicodeSPrintAsciiFormat(
		    &tmpStr[0],
		    TEMPSTRINGLENGTH,
		    tr.Filename);

		// 4. 文件名前面添加盘符
		StrCatS(DriveLetter,TEMPSTRINGLENGTH,tmpStr);

		if (DEBUGMODE)
		{
			Print(L"File would be saved as [%s]\n",DriveLetter);
		}

		// 如果该文件已经存在，那么删除现有的
		if (ShellFileExists(DriveLetter)==EFI_SUCCESS)
		{
			Print(L"Delete an existed file with same name\n");
			ShellOpenFileByName(
			    DriveLetter,
			    (SHELL_FILE_HANDLE *)&FileHandle,
			    EFI_FILE_MODE_CREATE|EFI_FILE_MODE_READ|EFI_FILE_MODE_WRITE,
			    0);
			ShellDeleteFile((SHELL_FILE_HANDLE *)&FileHandle);
		}
		if (DEBUGMODE)
		{
			Print(L"Create file\n");
		}
		
		// 创建接收文件
		Status = ShellOpenFileByName(
		             DriveLetter,
		             (SHELL_FILE_HANDLE *)&FileHandle,
		             EFI_FILE_MODE_CREATE|EFI_FILE_MODE_READ|EFI_FILE_MODE_WRITE,
		             0);
		if(Status != RETURN_SUCCESS)
		{
			Print(L"Create failed! %r\n",Status);
			return EFI_SUCCESS;
		}

		if (DEBUGMODE)
		{
			Print(L"Calc\n");
		}

		if (DEBUGMODE)
		{
			Print(L"File size ", tr.Size);
			Print(L"Counter", tr.Size / BlockIo->Media->BlockSize);
		}
		
		UINTN i,WriteSize;
		WriteSize=BlockIo->Media->BlockSize;
		for (i=0; i<tr.Size / BlockIo->Media->BlockSize; i++)
		{
			BlockIo->ReadBlocks(
			    BlockIo,
			    BlockIo->Media->MediaId,
			    (MAXDISKADDRESS - MAXPSRAMADDRESS)/(BlockIo->Media->BlockSize)+i,
			    BlockIo->Media->BlockSize,
			    Block);
			Status = ShellWriteFile(FileHandle,
			                        &WriteSize,
			                        Block);
		}
		// 如果当前的文件大小不是 BlockSize 的整数倍，那么需要特别处理一下
		if (tr.Size % BlockIo->Media->BlockSize!=0) {
			BlockIo->ReadBlocks(
			    BlockIo,
			    BlockIo->Media->MediaId,
			    (MAXDISKADDRESS - MAXPSRAMADDRESS)/(BlockIo->Media->BlockSize)+(tr.Size / BlockIo->Media->BlockSize),
			    BlockIo->Media->BlockSize,
			    Block);
			WriteSize=tr.Size % BlockIo->Media->BlockSize;
			if (DEBUGMODE)
			{
				Print(L"Write %d\n",WriteSize);
			}			
			Status = ShellWriteFile(FileHandle,
			                        &WriteSize,
			                        Block);
		}
		

		//Close the source file
		ShellCloseFile(&FileHandle);

		gBS->FreePool(BlockControllerHandles);
	}
	return EFI_SUCCESS;
}

