#include  <Uefi.h>
#include  <Library/UefiLib.h>
#include  <Library/ShellCEntryLib.h>
#include <Protocol/BlockIo.h>
#include <Library/MemoryAllocationLib.h>
#include <string.h>
#include <Library/ShellLib.h>
#include  <Library/PrintLib.h>

#define DEBUGMODE 0

extern EFI_BOOT_SERVICES         *gBS;
extern EFI_SYSTEM_TABLE			 *gST;
extern EFI_RUNTIME_SERVICES 	 *gRT;

// 用于  8MB PSRAM
#define MAXPSRAMADDRESS   8 * 1024 * 1024
// 用于 256MB Disk
#define MAXDISKADDRESS   256 * 1024 * 1024
		
UINT8 *SIGNATURE="DATTRANSFTERDISKWWW.LAB-Z.COM";
UINT8 *Block;

struct TranHeader
{
	UINT32 Flag;
	UINT32 Monot;
	UINT32 Size;
	UINT32 Counter;
	UINT16 FilenameLength;
	UINT8  Filename[494];
} tr;

INTN
EFIAPI
main (
    IN UINTN Argc,
    IN CHAR8 **Argv
)
{
	EFI_STATUS				Status;
	UINTN					HandleCount,HandleIndex;
	EFI_HANDLE              *BlockControllerHandles = NULL;
	EFI_BLOCK_IO_PROTOCOL   *BlockIo;
	BOOLEAN					Found=FALSE;
	EFI_FILE_HANDLE   		FileHandle;
	CHAR16					UnicodeBuffer[200];

	//找到全部有 BlockIo Protocol 的Device
	Status = gBS->LocateHandleBuffer(
	             ByProtocol,
	             &gEfiBlockIoProtocolGuid,
	             NULL,
	             &HandleCount,
	             &BlockControllerHandles);
	if (!EFI_ERROR(Status))
	{
		//逐个打开
		for (HandleIndex = 0; HandleIndex < HandleCount; HandleIndex++)
		{
			/*打开EFI_BLOCK_IO_PROTOCOL  */
			Status = gBS->HandleProtocol(
			             BlockControllerHandles[HandleIndex],
			             &gEfiBlockIoProtocolGuid,
			             (VOID**)&BlockIo);
			if (DEBUGMODE)
			{
				//显示信息
				Print(L"-->[Device]<--:%d\n",HandleIndex);
				Print(L"MediaId       :%0x\n",BlockIo->Media->MediaId);
				Print(L"RemovableMedia:%0x\n",BlockIo->Media->RemovableMedia);
				Print(L"MediaPresent  :%0x\n",BlockIo->Media->MediaPresent);
				Print(L"ReadOnly      :%0x\n",BlockIo->Media->ReadOnly);
				Print(L"WriteCaching  :%0x\n",BlockIo->Media->WriteCaching);
				Print(L"BlockSize     :%0x\n",BlockIo->Media->BlockSize);
				Print(L"IoAlign       :%0x\n",BlockIo->Media->IoAlign);
				Print(L"LastBlock     :%0x\n",BlockIo->Media->LastBlock);
				Print(L"LogicalPartition :%0x\n",BlockIo->Media->LogicalPartition);
				Print(L"LowestAlignedLba :%0x\n",BlockIo->Media->LowestAlignedLba);
				Print(L"LogicalBlocksPerPhysicalBlock   : %0x\n",
				      BlockIo->Media->LogicalBlocksPerPhysicalBlock);
				Print(L"OptimalTransferLengthGranularity: %0x\n",
				      BlockIo->Media->OptimalTransferLengthGranularity);
			}

			// 判断当前的 BlockIo 必须满足以下条件
			// 1. 必须是 RemovableMedia
			// 2. LastBlock 必须为0x7FFF
			if ((BlockIo->Media->RemovableMedia!=1)||(BlockIo->Media->LastBlock!=0x7FFFF))
			{
				if (DEBUGMODE)
				{
					Print(L"Break \n");
				}
				continue;
			}
			
			// 分配一个 BlockSize 大小的内存
			Block = AllocateZeroPool(BlockIo->Media->BlockSize);
			// 签名检查，用于判定设备是否为 DataTransfer
			BlockIo->ReadBlocks(
			    BlockIo,
			    BlockIo->Media->MediaId,
			    0x004, //0x800
			    BlockIo->Media->BlockSize,
			    Block);
			// 检查签名
			if (memcmp(SIGNATURE,Block,sizeof(SIGNATURE))!=0)
			{
				Print(L"Coundn't find Data Transfer\n");
				continue;
			}
			else
			{
				Found=TRUE;
				//Print(L"Found Data Transfer！\n");
				break;
			}

			FreePool(Block);
		}	//for (HandleIndex = 0;
		
		if (Found)
		{
			Print(L"We found DataTransfer Disk\n");
		} else {
			// 系统中无法找到 DataTransfer 设备
			Print(L"We coundn't find DataTransfer Disk in system\n");
			return EFI_SUCCESS;
		}

		// 检查 Busy 位
		BlockIo->ReadBlocks(
		    BlockIo,
		    BlockIo->Media->MediaId,
		    0x0C, //0x1800
		    BlockIo->Media->BlockSize,
		    Block);
		// 如果当前设备 Busy 中，那么退出
		if ((Block[0] & 1) == 0)
		{
			Print(L"Device busy, please try again later!\n");
			return EFI_SUCCESS;
		}
		// 如果命令行参数没有文件名, 报错推出
		if (Argc==0)
		{
			Print(L"Please input a file name\n");
			return EFI_SUCCESS;
		}
		else
		{
			if (DEBUGMODE)
			{
				Print(L"FilenameA [%a]\n",Argv[1]);
				Print(L"FilenameU [%s]\n",Argv[1]);
			}
		}

		// 将  Ascii 转为 Unicode
		UnicodeSPrintAsciiFormat(
		    &UnicodeBuffer[0],
		    sizeof(UnicodeBuffer),
		    Argv[1]);
			
		Status = ShellFileExists(UnicodeBuffer);
		if (Status!=EFI_SUCCESS) {
			Print(L"Can't find %s\n",UnicodeBuffer);
			return EFI_SUCCESS;
		}
		
		// 打开要发送的文件
		Status = ShellOpenFileByName(
		             UnicodeBuffer,
		             (SHELL_FILE_HANDLE *)&FileHandle,
		             EFI_FILE_MODE_READ,
		             0);
		if(Status != RETURN_SUCCESS)
		{
			Print(L"OpenFile failed!\n");
			return EFI_SUCCESS;
		}
		
		// 取得文件长度
		EFI_FILE_INFO     *FileInfo = NULL;
		FileInfo = ShellGetFileInfo( (SHELL_FILE_HANDLE)FileHandle);
		if (DEBUGMODE)
		{
			Print(L"Filesize [%ld] bytes\n",FileInfo-> FileSize);
			Print(L"PhysicalSize [%ld] bytes\n",FileInfo-> PhysicalSize);
		}
		
		//取文件长度
		tr.Size = (UINT32)FileInfo-> FileSize;
		//文件数量
		tr.Counter = 0;

		// 赋值文件名
		memcpy(tr.Filename,Argv[1],AsciiStrLen(Argv[1]));
		Print(L"FilenameA [%a]\n",tr.Filename);
		
		//文件名长度
		tr.FilenameLength = (UINT16)AsciiStrLen(Argv[1]);

		tr.Flag = (UINT32)(Block[0] + (Block[1] << 8) + (Block[2] << 16) + (Block[3] << 24));
		tr.Monot = (UINT32)(Block[4] + (Block[5] << 8) + (Block[6] << 16) + (Block[7] << 24));
		// 自增标志加一
		tr.Monot++;
		// 设置 Busy位
		tr.Flag = tr.Flag & 0xFFFFFFFE;
		// 回写，表明现在正在写入操作
		BlockIo->WriteBlocks(
		    BlockIo,
		    BlockIo->Media->MediaId,
		    0x0C, //0x800
		    BlockIo->Media->BlockSize,
		    &tr);

		// 要循环写入的次数
		UINT32 Counter;
		if (tr.Size%BlockIo->Media->BlockSize==0) {
			Counter=tr.Size / BlockIo->Media->BlockSize;
		} else {
			Counter=(tr.Size / BlockIo->Media->BlockSize) +1;
		}
		UINTN i,ReadSize;
		ReadSize=BlockIo->Media->BlockSize;
		for (i=0;i<Counter;i++)
		{
			Status = ShellReadFile(FileHandle,
			                       &ReadSize,
			                       Block);
			BlockIo->WriteBlocks(
			    BlockIo,
			    BlockIo->Media->MediaId,
			    (MAXDISKADDRESS - MAXPSRAMADDRESS)/(BlockIo->Media->BlockSize)+i, 
			    BlockIo->Media->BlockSize,
			    Block);
		}

		//清除 Busy 位
		tr.Flag = tr.Flag | 0x1;
		BlockIo->WriteBlocks(
		    BlockIo,
		    BlockIo->Media->MediaId,
		    0x0C, //0x800
		    BlockIo->Media->BlockSize,
		    &tr);

		//Close the source file
		ShellCloseFile(&FileHandle);

		gBS->FreePool(BlockControllerHandles);
	}
	return EFI_SUCCESS;
}

